#include <unistd.h>
#include <cstring>
#include <cerrno>
#include <sys/wait.h>
#include <sys/prctl.h>

#include "log.h"
#include "memory.h"
#include "loop.h"
#include "signals.h"

// Main
int main(__attribute__((unused)) int argc, char *argv[]) {
    // Check Arguments
    if (argc < 2) {
        ERR("Invalid Arguments");
    }
    // Setup
    create_pipes();
    // Fork
    pid_t pid = fork();
    if (pid == -1) {
        ERR("Unable To Fork Process: %s", strerror(errno));
    } else if (pid == 0) {
        // Setup
        setpgid(0, 0);
        init_pipe_guest();
        // Kill Child If Parent Exits First
        prctl(PR_SET_PDEATHSIG, SIGKILL);
        // Execute Program
        execvp(argv[1], (char *const *) &argv[1]);
        ERR("Unable To Execute Program: %s: %s", argv[1], strerror(errno));
    } else {
        // Parent

        // Setup Trampoline
        init_signals(pid);
        init_memory(pid);

        // Read From Pipes
        pipe_read_loop();

        // Reap Child
        int status;
        if (waitpid(pid, &status, 0) == -1) {
            ERR("Unable To Reap Child: %s", strerror(errno));
        }
        // Exit
        if (WIFEXITED(status)) {
            exit(WEXITSTATUS(status));
        } else {
            ERR("Unusual Exit");
        }
    }
}